#!/usr/bin/env node

/*************************************************************************/
/*    Copyright (C) 2017 Wizardry and Steamworks - License: GNU GPLv3    */
/*************************************************************************/

const url = require('url');
const path = require('path');
const fs = require('fs');
const mime = require('mime');
const auth = require("http-auth");
const stream = require('stream');

// Checks whether userPath is a child of rootPath.
function isRooted(userPath, rootPath, separator, callback) {
	userPath = userPath.split(separator).filter(Boolean);
	rootPath = rootPath.split(separator).filter(Boolean);
	callback(userPath.length >= rootPath.length &&
		rootPath.every((e, i) => e === userPath[i]));
}

// Serves files.
function files(config, file, client, callback) {
	// Check if the file is accessible.
	fs.access(file, fs.constants.R_OK, (error) => {
		if (error) {
			callback({
				message: 'Client: ' +
					client.address + ':' +
					client.port +
					' requesting inaccessible path: ' +
					file,
				severity: 'warning'
			}, {
				status: 403,
				data: new stream.Readable({
					read(size) {
						this.push(null);
					}
				}),
				type: 'text/plain'
			});
			return;
		}
		callback({
			message: 'Client: ' +
				client.address + ':' +
				client.port +
				' sent file: ' +
				file,
			severity: 'info'
		}, {
			status: 200,
			data: fs
				.createReadStream(file),
			type: mime
				.lookup(file)
		});
	});
}

// Serves a directory listing or the document index in case it exists.
function index(config, directory, href, client, callback) {
	const root = path.resolve(directory, config.site.index);
	fs.stat(root, (error, stats) => {
		if (error) {
			if (config.site.indexing
				.some((directory) =>
					directory.toUpperCase() === href.toUpperCase())) {
				fs.readdir(directory, (error, paths) => {
					if (error) {
						console.log("listing forbidden...");
						callback({
							message: 'Client: ' +
								client.address + ':' +
								client.port +
								' could not access directory: ' +
								directory,
							severity: 'warning'
						}, {
							status: 500,
							data: new stream.Readable({
								read(size) {
									this.push(null);
								}
							}),
							type: 'text/plain'
						});
						return;
					}
					console.log("sending listing...");
					callback({
						message: 'Client: ' +
							client.address + ':' +
							client.port +
							' accessed directory listing: ' +
							directory,
						severity: 'info'
					}, {
						status: 200,
						data: new stream.Readable({
							read(size) {
								this.push(JSON.stringify(paths));
								this.push(null);
							}
						}),
						type: 'application/json'
					});
				});
				return;
			}
			// Could not access directory index file and directory listing not allowed.
			console.log("no dirindex...");
			callback({
				message: 'Client: ' +
					client.address + ':' +
					client.port +
					' no index file found and accessing forbiden index: ' +
					href,
				severity: 'warning'
			}, {
				status: 403,
				data: new stream.Readable({
					read(size) {
						this.push(null);
					}
				}),
				type: 'text/plain'
			});
			return;
		}

		// Serve the document index.
		fs.access(root, fs.constants.R_OK, (error) => {
			if (error) {
				callback({
					message: 'Client: ' +
						client.address + ':' +
						client.port +
						' unable to access path: ' +
						directory,
					severity: 'warning'
				}, {
					status: 403,
					data: new stream.Readable({
						read(size) {
							this.push(null);
						}
					}),
					type: 'text/plain'
				});
				return;
			}
			callback({
				message: 'Client: ' +
					client.address + ':' +
					client.port +
					' sent file: ' +
					root,
				severity: 'info'
			}, {
				status: 200,
				data: fs.createReadStream(root),
				type: mime.lookup(root)
			});
		});
	});
}

// Determines whether the requested filesystem request path is a directory or a file.
function serve(config, local, href, address, callback) {
	fs.stat(local, (error, stats) => {
		// Document does not exist.
		if (error) {
			callback({
				message: 'Client: ' +
					address.address + ':' +
					address.port +
					' accessing non-existent document: ' +
					local,
				severity: 'warning'
			}, {
				status: 404,
				data: new stream.Readable({
					read(size) {
						this.push(null);
					}
				}),
				type: 'text/plain'
			});
			return;
		}

		if (stats.isDirectory()) {
			// Directory is requested so provide directory indexes.
			index(config, local, href, address, callback);
			return;
		}
		if (stats.isFile()) {
			const file = path.parse(local).base;

			// If the file matches the reject list or is not in the accept list,
			// then there is no file to serve. 
			if (config.site.reject.some((expression) => expression.test(file)) ||
				!config.site.accept.some((expression) => expression.test(file))) {
				callback({
					message: 'Client: ' +
						address.address + ':' +
						address.port +
						' requested disallowed file: ' +
						file,
					severity: 'warning'
				}, {
					status: 404,
					data: new stream.Readable({
						read(size) {
							this.push(null);
						}
					}),
					type: 'text/plain'
				});
				return;
			}

			// A file was requested so provide the file.
			files(config, local, address, callback);
		}
	});
}

module.exports = {
	process: (config, request, response, root, callback) => {
		const requestAddress = request.socket.address();
		const requestURL = url.parse(
			request.url, true
		);

		// Perform URL re-writes.
		Object.keys(config.site.rewrite).forEach((key, index) => {
			if (config.site.rewrite[key].test(requestURL.path)) {
				const originalPath = requestURL.path;
				requestURL.path = requestURL
					.path
					.replace(
						config.site.rewrite[key], key
					);
				requestURL.pathname = url.parse(
						requestURL
						.pathname
						.replace(
							config.site.rewrite[key], key
						),
						true
					)
					.pathname;
				callback({
					message: 'Rewrite path: ' +
						originalPath +
						' to: ' +
						requestURL.path,
					severity: 'info'
				});
			}
		});

		const trimmedPath = requestURL
			.pathname
			.split('/')
			.filter(Boolean)
			.join('/');
		const requestPath = trimmedPath === '/' ?
			path.join(root, trimmedPath) :
			path.resolve(root, trimmedPath);

		fs.realpath(requestPath, (error, resolvedPath) => {
			// If the path does not exist, then return early.
			if (error) {
				callback({
					message: 'Unknown path requested: ' +
						requestAddress.address + ':' +
						requestAddress.port +
						' requesting: ' +
						requestURL.pathname,
					severity: 'warning'
				}, {
					status: 404,
					data: new stream.Readable({
						read(size) {
							this.push(null);
						}
					}),
					type: 'text/plain'
				});
				return;
			}
			// Check for path traversals early on and bail if the requested path does not
			// lie within the specified document root.
			isRooted(resolvedPath, root, path.sep, (rooted) => {
				if (!rooted) {
					callback({
						message: 'Attempted path traversal: ' +
							requestAddress.address + ':' +
							requestAddress.port +
							' requesting: ' +
							requestURL.pathname,
						severity: 'warning'
					}, {
						status: 404,
						data: new stream.Readable({
							read(size) {
								this.push(null);
							}
						}),
						type: 'text/plain'
					});
					return;
				}

				// If authentication is required for this path then perform authentication.
				if (config.auth.locations.some(
						(authPath) => authPath.toUpperCase() === requestURL.pathname.toUpperCase())) {
					// Create digest authentication.
					const authentication = auth.digest({
						realm: config.auth.realm,
						file: path.resolve(
							path.dirname(require.main.filename),
							config.auth.digest
						)
					});
					// Requested location requires authentication.
					authentication.check(request, response, (request, response) => {
						callback({
							message: 'Authenticated client: ' +
								requestAddress.address + ':' +
								requestAddress.port +
								' accessing: ' +
								requestURL.pathname,
							severity: 'info'
						});
						serve(config,
							requestPath,
							requestURL.pathname,
							requestAddress,
							callback
						);
					});
					return;
				}

				// If no authentication is required then serve the request.
				callback({
					message: 'Client: ' +
						requestAddress.address + ':' +
						requestAddress.port +
						' accessing: ' +
						requestURL.pathname,
					severity: 'info'
				});
				serve(config,
					requestPath,
					requestURL.pathname,
					requestAddress,
					callback
				);
			});
		});
	}
};
